# -*- coding: utf-8 -*-
"""Project metadata

Information describing the project.
"""

# The package name, which is also the "UNIX name" for the project.
package = 'piecash'
project = "Python interface to GnuCash documents"
project_no_spaces = project.replace(' ', '')
version = '0.4.4'
description = 'A pythonic interface to GnuCash SQL documents.'
authors = ['sdementen']
authors_string = ', '.join(authors)
emails = ['sdementen@gmail.com']
license = 'MIT'
copyright = '2014 ' + authors_string
url = 'https://github.com/sdementen/piecash'
